"""
Tray icon for udiskie.
"""
__all__ = ['create_menu',
           'create_statusicon',
           'connect_statusicon',
           'main']

import gtk
from functools import partial
from collections import namedtuple
from itertools import chain


def setdefault(self, other):
    """Merge two dictionaries like .update() but don't overwrite values."""
    for k,v in other.items():
        self.setdefault(k, v)


TreeNode = namedtuple('TreeNode',
                      ['root', 'branches',
                       'device', 'label', 'methods'])
Action = namedtuple('Action', ['label', 'device', 'method'])
Branch = namedtuple('Branch', ['label', 'groups'])


def device_tree(udisks):
    """
    Return the device hierarchy as list of TreeNodes
    """
    devices = {}
    rootnode = TreeNode(None, [], None, "", [])

    def find_node(object_path):
        if object_path in devices:
            return devices[object_path]
        else:
            return mknode(udisks.create_device(object_path))

    def mknode(device):
        # methods
        methods = []
        label = device.device_file
        if device.is_filesystem:
            if device.is_mounted:
                methods.append('unmount')
                label = device.mount_paths[0]
            else:
                methods.append('mount')
        elif device.is_crypto:
            if device.is_unlocked:
                methods.append('lock')
            else:
                methods.append('unlock')
        if device.is_ejectable:
            methods.append('eject')
        if device.is_detachable:
            methods.append('detach')

        # root
        if device.is_partition:
            root = find_node(device.partition_slave)
        elif device.is_luks_cleartext:
            root = find_node(device.luks_cleartext_slave)
        else:
            root = rootnode

        node = TreeNode(root, [], device, label, methods)
        devices[device.object_path] = node
        root.branches.append(node)
        return node

    for device in udisks.get_all_handleable():
        if device.object_path not in devices:
            mknode(device)

    return rootnode

def simple_menu(node):
    return Branch(
        label=node.label,
        groups=[
            [flatten_menu(branch)
             for branch in node.branches],
            [Action(node.label, node.device, method)
             for method in node.methods],
        ])

def flat_menu(node):
    def actions(node, presentation):
        return [Action(presentation, node.device, method)
                for method in node.methods]

    def leaves(node, outer_methods, presentation):
        if not presentation or (node.device.is_mounted or
                                not node.device.is_luks_cleartext):
            presentation = node.label
        if node.branches:
            return chain.from_iterable(
                leaves(branch,
                       actions(node, presentation) + outer_methods,
                       presentation)
                for branch in node.branches)
        elif len(node.methods) + len(outer_methods) > 0:
            return Branch(
                label=presentation,
                groups=[list(chain(actions(node,
                                           presentation),
                                   outer_methods))]),
        else:
            return ()

    return Branch(
        label=node.label,
        groups=[list(leaves(node, [], ""))])


def create_menu(udisks=None,
                mounter=None,
                labels={},
                icons={},
                actions={},
                style=flat_menu,
                flat=False):
    """
    Create menu for udiskie mount operations.

    :param object udisks: Interface to udisks used to iterate external devices
    :param object mounter: Mount operation provider
    :param dict labels: Labels for menu items
    :param dict icons: Icons for menu items
    :param dict actions: Actions for menu items
    :param bool flat: Create a flattened menu

    If either ``udisks`` and or ``mounter`` is ``None`` default versions
    will be imported from the udiskie package.

    Valid keys for the ``labels``, ``icons`` and ``actions`` dictionaries are:

        - mount     Mount a device
        - unmount   Unmount a device
        - unlock    Unlock a LUKS device
        - lock      Lock a LUKS device
        - eject     Eject a drive
        - detach    Detach (power down) a drive
        - quit      Exit the application

    NOTE: If using a main loop other than ``gtk.main`` the 'quit' action
    must be customized.

    To prevent a certain action from being displayed its ``action`` must be
    set to ``None``.

    """
    if udisks is None:
        from dbus import SystemBus
        from udiskie.udisks import Udisks
        udisks = Udisks.create(SystemBus())
    if mounter is None:
        from udiskie.mount import Mounter
        from udiskie.prompt import password
        mounter = Mounter(prompt=password(), udisks=udisks)

    setdefault(icons, {
        'mount': gtk.STOCK_APPLY,
        'unmount': gtk.STOCK_CANCEL,
        'unlock': gtk.STOCK_APPLY,
        'lock': gtk.STOCK_CANCEL,
        'eject': gtk.STOCK_CANCEL,
        'detach': gtk.STOCK_CANCEL,
        'quit': gtk.STOCK_QUIT, })

    setdefault(labels, {
        'mount': 'Mount %s',
        'unmount': 'Unmount %s',
        'unlock': 'Unlock %s',
        'lock': 'Lock %s',
        'eject': 'Eject %s',
        'detach': 'Detach %s',
        'quit': 'Quit', })

    setdefault(actions, {
        'mount': mounter.mount_device,
        'unmount': mounter.unmount_device,
        'unlock': mounter.unlock_device,
        'lock': partial(mounter.remove_device, force=True),
        'eject': partial(mounter.eject_device, force=True),
        'detach': partial(mounter.detach_device, force=True),
        'quit': gtk.main_quit, })

    def create_menuitem(label, icon, onclick):
        if icon is None:
            item = gtk.MenuItem()
        else:
            try:
                item = gtk.ImageMenuItem(stock_id=icon)
            except TypeError:
                item = gtk.ImageMenuItem()
                item.set_icon(icon)
        if label is not None:
            item.set_label(label)
        if isinstance(onclick, gtk.Menu):
            item.set_submenu(onclick)
        else:
            item.connect('activate', onclick)
        return item

    def item(action, feed=(), bind=()):
        return create_menuitem(
            labels[action] % tuple(feed),
            icons[action],
            lambda _: actions[action](*bind))

    def mkmenu(menu_node):
        menu = gtk.Menu()
        separate = False
        for group in menu_node.groups:
            if len(group) > 0:
                if separate:
                    menu.append(gtk.SeparatorMenuItem())
                separate = True
            for node in group:
                if isinstance(node, Action):
                    menu.append(item(
                        node.method,
                        feed=[node.label],
                        bind=[node.device]))
                else:
                    menu.append(create_menuitem(
                        node.label,
                        icon=None,
                        onclick=mkmenu(node)))
        return menu

    # create udisks actions items
    menu = mkmenu(flat_menu(device_tree(udisks)))

    # append menu item for closing the application
    if actions['quit']:
        if len(menu) > 0:
            menu.append(gtk.SeparatorMenuItem())
        menu.append(item('quit'))

    return menu


def create_statusicon():
    """Create a simple gtk.StatusIcon"""
    statusicon = gtk.StatusIcon()
    statusicon.set_from_stock(gtk.STOCK_CDROM)
    statusicon.set_tooltip("udiskie")
    return statusicon


def connect_statusicon(statusicon, menu=create_menu):
    """Connect a popup menu event handler, return the connection identifier."""
    def right_click_event(icon, button, time):
        m = menu()
        m.show_all()
        m.popup(parent_menu_shell=None,
                parent_menu_item=None,
                func=gtk.status_icon_position_menu,
                button=button,
                activate_time=time,
                data=icon)
    return statusicon.connect("popup-menu", right_click_event)


def main():
    """Run udiskie tray icon in a main loop."""
    statusicon = create_statusicon()
    connection = connect_statusicon(statusicon)
    gtk.main()
    statusicon.disconnect(connection)
    statusicon.set_visible(False)


if __name__ == '__main__':
    main()




"""
Udisks wrapper utilities.

These act as a convenience abstraction layer on the udisks dbus service.
Requires Udisks 1.0.5 as described here:

    http://udisks.freedesktop.org/docs/1.0.5/

Note that (as this completely wraps the udisks dbus API) replacing this
module will let you support Udisks2 or maybe even other services.

"""
__all__ = ['Device', 'Udisks']

import logging
import os
from udiskie.common import DBusProxy


UDISKS_INTERFACE = 'org.freedesktop.UDisks'
UDISKS_DEVICE_INTERFACE = 'org.freedesktop.UDisks.Device'

UDISKS_OBJECT = 'org.freedesktop.UDisks'
UDISKS_OBJECT_PATH = '/org/freedesktop/UDisks'


class Device(DBusProxy):
    """
    Wrapper class for org.freedesktop.UDisks.Device proxy objects.
    """
    # construction
    def __init__(self, bus, proxy):
        """
        Initialize an instance with the given dbus proxy object.

        proxy must be an object acquired by a call to bus.get_object().

        """
        self.log = logging.getLogger('udiskie.udisks.Device')
        super(Device, self).__init__(proxy, UDISKS_DEVICE_INTERFACE)
        self.bus = bus

    @classmethod
    def create(cls, bus, object_path):
        """Find the object on the specified bus."""
        return cls(bus, bus.get_object(UDISKS_OBJECT, object_path))

    # string representation
    def __str__(self):
        return self.object_path

    def __eq__(self, other):
        if isinstance(other, Device):
            return self.object_path == other.object_path
        else:
            return self.object_path == str(other)

    def __ne__(self, other):
        return not (self == other)

    # check if the device is a valid udisks object
    @property
    def is_valid(self):
        try:
            self.property.DeviceFile
            return True
        except self.Exception:
            return False

    # properties
    @property
    def partition_slave(self):
        """Get the partition slave (container)."""
        return self.property.PartitionSlave if self.is_partition else None

    @property
    def is_partition(self):
        """Check if the device has a partition slave."""
        return self.property.DeviceIsPartition

    @property
    def is_partition_table(self):
        """Check if the device is a partition table."""
        return self.property.DeviceIsPartitionTable

    @property
    def is_drive(self):
        """Check if the device is a drive."""
        return self.property.DeviceIsDrive

    @property
    def drive(self):
        """
        Get the drive containing this device.

        The returned Device object is not guaranteed to be a drive.

        """
        if self.is_partition:
            return self.create(self.bus, self.partition_slave).drive
        elif self.is_luks_cleartext:
            return self.create(self.bus, self.luks_cleartext_slave).drive
        else:
            return self

    @property
    def is_detachable(self):
        """Check if the drive that owns this device can be detached."""
        return self.property.DriveCanDetach if self.is_drive else None

    @property
    def is_ejectable(self):
        """Check if the drive that owns this device can be ejected."""
        return self.property.DriveIsMediaEjectable if self.is_drive else None

    @property
    def is_systeminternal(self):
        """Check if the device is internal."""
        return self.property.DeviceIsSystemInternal

    @property
    def is_external(self):
        """Check if the device is external."""
        return not self.is_systeminternal

    @property
    def is_handleable(self):
        """
        Should this device be handled by udiskie?

        Currently this just means that the device is removable and holds a
        filesystem or the device is a LUKS encrypted volume.

        """
        return (self.is_filesystem or self.is_crypto) and self.is_external

    @property
    def is_mounted(self):
        """Check if the device is mounted."""
        return self.property.DeviceIsMounted

    @property
    def is_unlocked(self):
        """Check if device is already unlocked."""
        return self.luks_cleartext_holder if self.is_luks else None

    @property
    def mount_paths(self):
        """Return list of active mount paths."""
        if not self.is_mounted:
            return []
        raw_paths = self.property.DeviceMountPaths
        return [os.path.normpath(path) for path in raw_paths]

    @property
    def device_file(self):
        """The filesystem path of the device block file."""
        return os.path.normpath(self.property.DeviceFile)

    @property
    def is_filesystem(self):
        return self.id_usage == 'filesystem'

    @property
    def is_crypto(self):
        return self.id_usage == 'crypto'

    @property
    def is_luks(self):
        return self.property.DeviceIsLuks

    @property
    def is_luks_cleartext(self):
        """Check whether this is a luks cleartext device."""
        return self.property.DeviceIsLuksCleartext

    @property
    def luks_cleartext_slave(self):
        """Get luks crypto device."""
        return self.property.LuksCleartextSlave if self.is_luks_cleartext else None

    @property
    def luks_cleartext_holder(self):
        """Get unlocked luks cleartext device."""
        return self.property.LuksHolder if self.is_luks else None

    @property
    def is_luks_cleartext_slave(self):
        """Check whether the luks device is currently in use."""
        if not self.is_luks:
            return False
        for device in Udisks.create(self.bus).get_all():
            if (not device.is_filesystem or device.is_mounted) and (
                    device.is_luks_cleartext and
                    device.luks_cleartext_slave == self.object_path):
                return True
        return False

    @property
    def has_media(self):
        return self.property.DeviceIsMediaAvailable

    @property
    def id_type(self):
        return self.property.IdType

    @property
    def id_usage(self):
        return self.property.IdUsage

    @property
    def id_uuid(self):
        """Device UUID."""
        return self.property.IdUuid

    # methods
    def mount(self, filesystem=None, options=[]):
        """Mount filesystem."""
        if filesystem is None:
            filesystem = self.id_type
        self.method.FilesystemMount(filesystem, options)

    def unmount(self, options=[]):
        """Unmount filesystem."""
        self.method.FilesystemUnmount(options)

    def lock(self, options=[]):
        """Lock Luks device."""
        return self.method.LuksLock(options)

    def unlock(self, password, options=[]):
        """Unlock Luks device."""
        return self.method.LuksUnlock(password, options)

    def eject(self, options=[]):
        """Eject media from the device."""
        return self.method.DriveEject(options)

    def detach(self, options=[]):
        """Detach the device by e.g. powering down the physical port."""
        return self.method.DriveDetach(options)


class Udisks(DBusProxy):
    """
    Udisks dbus service wrapper.

    This is a dbus proxy object to the org.freedesktop.UDisks interface of
    the udisks service object.

    """
    # Construction
    def __init__(self, bus, proxy):
        """
        Initialize an instance with the given dbus proxy object.

        proxy must be an object acquired by a call to bus.get_object().

        """
        super(Udisks, self).__init__(proxy, UDISKS_INTERFACE)
        self.bus = bus

    @classmethod
    def create(cls, bus):
        """Connect to the udisks service on the specified bus."""
        return cls(bus, bus.get_object(UDISKS_OBJECT, UDISKS_OBJECT_PATH))

    # instantiation of device objects
    def create_device(self, object_path):
        """Create a Device instance from object path."""
        return Device.create(self.bus, object_path)

    # Methods
    def get_all(self):
        """
        Enumerate all device objects currently known to udisks.

        NOTE: returns only devices that are still valid. This protects from
        race conditions inside udiskie.

        """
        for object_path in self.method.EnumerateDevices():
            dev = self.create_device(object_path)
            if dev.is_valid:
                yield dev

    def get_all_handleable(self):
        """
        Enumerate all handleable devices currently known to udisks.

        NOTE: returns only devices that are still valid. This protects from
        race conditions inside udiskie.

        """
        return (dev for dev in self.get_all() if dev.is_handleable)

    def get_device(self, path):
        """
        Get a device proxy by device name or any mount path of the device.

        This searches through all accessible devices and compares device
        path as well as mount pathes.

        """
        logger = logging.getLogger('udiskie.udisks.get_device')
        for device in self.get_all():
            if os.path.samefile(path, device.device_file):
                return device
            for p in device.mount_paths:
                if os.path.samefile(path, p):
                    return device
        logger.warn('Device not found: %s' % path)
        return None




"""
Udiskie notification daemon.
"""
__all__ = ['Notify']

class Notify(object):
    """
    Notification tool.

    Can be connected to udisks daemon in order to automatically issue
    notifications when system status has changed.

    """
    def __init__(self, notify):
        """
        Initialize notifier.

        A notify service such as pynotify or notify2 should be passed in.

        """
        self.notify = notify

    # event handlers:
    def device_mounted(self, device):
        device_file = device.device_file
        mount_path = device.mount_paths[0]
        self.notify.Notification(
            'Device mounted',
            '%s mounted on %s' % (device_file, mount_path),
            'drive-removable-media').show()

    def device_unmounted(self, device):
        device_file = device.device_file
        self.notify.Notification(
            'Device unmounted',
            '%s unmounted' % (device_file,),
            'drive-removable-media').show()

    def device_locked(self, device):
        device_file = device.device_file
        self.notify.Notification(
            'Device locked',
            '%s locked' % (device_file,),
            'drive-removable-media').show()

    def device_unlocked(self, device):
        device_file = device.device_file
        self.notify.Notification(
            'Device unlocked',
            '%s unlocked' % (device_file,),
            'drive-removable-media').show()




"""
Common DBus utilities.
"""
__all__ = ['DBusProperties', 'DBusProxy']

from dbus import Interface
from dbus.exceptions import DBusException

class DBusProperties(object):
    """
    Dbus property map abstraction.

    Properties of the object can be accessed as attributes.

    """
    def __init__(self, dbus_object, interface):
        """Initialize a proxy object with standard dbus property interface."""
        self.__proxy = Interface(
                dbus_object,
                dbus_interface='org.freedesktop.DBus.Properties')
        self.__interface = interface

    def __getattr__(self, property):
        """Retrieve the property via the dbus proxy."""
        return self.__proxy.Get(self.__interface, property)

class DBusProxy(object):
    """
    DBus proxy object.

    Provides property and method bindings.

    """
    def __init__(self, proxy, interface):
        self.Exception = DBusException
        self.proxy = proxy
        self.object_path = proxy.object_path
        self.property = DBusProperties(self.proxy, interface)
        self.method = Interface(self.proxy, interface)




"""
Udiskie user prompt utility.
"""
__all__ = ['password']

import subprocess
from distutils.spawn import find_executable

def password(prompt_name='zenity'):
    """
    Create a password prompt function.
    """
    if not prompt_name:
        return None

    executable = find_executable(prompt_name)
    if executable is None:
        return None

    # builtin variant: enter password via zenity:
    if prompt_name == 'zenity':
        def doprompt(text, title):
            return subprocess.check_output([executable,
                '--entry', '--hide-text',
                '--text', text, '--title', title ])

    # builtin variant: enter password via systemd-ask-password:
    elif prompt_name == 'systemd-ask-password':
        def doprompt(text, title):
            return subprocess.check_output([executable, text])

    # enter password via user supplied binary:
    else:
        def doprompt(text, title):
            return subprocess.check_output([executable, text, title])

    def password_prompt(text, title):
        try:
            return doprompt(text, title).decode('utf-8').rstrip('\n')
        except subprocess.CalledProcessError:
            return None

    return password_prompt



"""
Udiskie CLI logic.
"""
__all__ = [
    # entry points:
    'daemon',
    'mount',
    'umount',
    ]

import warnings
warnings.filterwarnings("ignore", ".*could not open display.*", Warning)
warnings.filterwarnings("ignore", ".*g_object_unref.*", Warning)

import os
import logging
from functools import partial

CONFIG_PATH = 'udiskie/filters.conf'

#----------------------------------------
# Utility functions
#----------------------------------------
def load_filter(filter_file=None):
    """Load mount option filters."""
    import udiskie.match
    if not filter_file:
        try:
            from xdg.BaseDirectory import xdg_config_home
        except ImportError:
            xdg_config_home = os.path.expanduser('~/.config')
        filter_file = os.path.join(xdg_config_home, CONFIG_PATH)
    return udiskie.match.FilterMatcher((filter_file,))

def common_program_options():
    """
    Return a command line option parser for options common to all modes.
    """
    import optparse
    parser = optparse.OptionParser()
    parser.add_option('-v', '--verbose', action='store_const',
                      dest='log_level', default=logging.INFO,
                      const=logging.DEBUG, help='verbose output')
    return parser

def mount_program_options():
    """
    Return the mount option parser for the mount command.
    """
    parser = common_program_options()
    parser.add_option('-f', '--filters', action='store',
                      dest='filters', default=None,
                      metavar='FILE', help='filter FILE')
    parser.add_option('-P', '--password-prompt', action='store',
                      dest='password_prompt', default='zenity',
                      metavar='MODULE', help="replace password prompt")
    return parser

def connect_udisks(bus):
    """
    Return a connection to the first udisks service found available.

    TODO: This should check if the udisks service is accessible and if not
    try to connect to udisks2 service.

    """
    import udiskie.udisks
    return udiskie.udisks.Udisks.create(bus)


#----------------------------------------
# Entry points
#----------------------------------------
def daemon(args=None, udisks=None):
    """
    Execute udiskie as a daemon.
    """
    import gobject
    import udiskie.automount
    import udiskie.daemon
    import udiskie.mount
    import udiskie.prompt

    parser = mount_program_options()
    parser.add_option('-s', '--suppress', action='store_true',
                      dest='suppress_notify', default=False,
                      help='suppress popup notifications')
    parser.add_option('-t', '--tray', action='store_true',
                      dest='tray', default=False,
                      help='show tray icon')
    options, posargs = parser.parse_args(args)
    logging.basicConfig(level=options.log_level, format='%(message)s')

    # for now: just use the default udisks
    if udisks is None:
        import dbus
        from dbus.mainloop.glib import DBusGMainLoop
        DBusGMainLoop(set_as_default=True)
        bus = dbus.SystemBus()
        udisks = connect_udisks(bus)

    # create daemon
    daemon = udiskie.daemon.Daemon(udisks=udisks)
    mainloop = gobject.MainLoop()

    # create a mounter
    prompt = udiskie.prompt.password(options.password_prompt)
    filter = load_filter(options.filters)
    mounter = udiskie.mount.Mounter(filter=filter, prompt=prompt, udisks=udisks)

    # notifications (optional):
    if not options.suppress_notify:
        import udiskie.notify
        try:
            import notify2 as notify_service
        except ImportError:
            import pynotify as notify_service
        notify_service.init('udiskie.mount')
        notify = udiskie.notify.Notify(notify_service)
        daemon.connect(notify)

    # tray icon (optional):
    if options.tray:
        import udiskie.tray
        create_menu = partial(udiskie.tray.create_menu,
                              udisks=udisks,
                              mounter=mounter,
                              actions={'quit': mainloop.quit})
        statusicon = udiskie.tray.create_statusicon()
        connection = udiskie.tray.connect_statusicon(statusicon, create_menu)

    # automounter
    automount = udiskie.automount.AutoMounter(mounter)
    daemon.connect(automount)

    mounter.mount_all()
    try:
        return mainloop.run()
    except KeyboardInterrupt:
        return 0

def mount(args=None, udisks=None):
    """
    Execute the mount command.
    """
    import udiskie.mount
    import udiskie.prompt

    parser = mount_program_options()
    parser.add_option('-a', '--all', action='store_true',
                      dest='all', default=False,
                      help='mount all present devices')
    options, posargs = parser.parse_args(args)
    logging.basicConfig(level=options.log_level, format='%(message)s')

    # for now: just use the default udisks
    if udisks is None:
        import dbus
        bus = dbus.SystemBus()
        udisks = connect_udisks(bus)

    # create a mounter
    prompt = udiskie.prompt.password(options.password_prompt)
    filter = load_filter(options.filters)
    mounter = udiskie.mount.Mounter(filter=filter, prompt=prompt, udisks=udisks)

    # mount all present devices
    if options.all:
        mounter.mount_all()
        return 0

    # only mount the desired devices
    elif len(posargs) > 0:
        mounted = []
        for path in posargs:
            device = mounter.mount(path)
            if device:
                mounted.append(device)
        # automatically mount luks holders
        for device in mounted:
            mounter.mount_holder(device)
        return 0

    # print command line options
    else:
        parser.print_usage()
        return 1

def umount(args=None, udisks=None):
    """
    Execute the umount command.
    """
    import udiskie.mount

    parser = common_program_options()
    parser.add_option('-a', '--all', action='store_true',
                      dest='all', default=False,
                      help='all devices')
    parser.add_option('-e', '--eject', action='store_true',
                      dest='eject', default=False,
                      help='Eject drive')
    parser.add_option('-d', '--detach', action='store_true',
                      dest='detach', default=False,
                      help='Detach drive')
    (options, posargs) = parser.parse_args(args)
    logging.basicConfig(level=options.log_level, format='%(message)s')

    if len(posargs) == 0 and not options.all:
        parser.print_usage()
        return 1

    # for now: use udisks v1 service
    if udisks is None:
        import dbus
        bus = dbus.SystemBus()
        udisks = connect_udisks(bus)
    mounter = udiskie.mount.Mounter(udisks=udisks)

    unmounted = []
    if options.all:
        if options.eject or options.detach:
            if options.eject:
                mounter.eject_all()
            if options.detach:
                mounter.detach_all()
        else:
            unmounted = mounter.unmount_all()
    else:
        unmounted = []
        for path in posargs:
            if options.eject or options.detach:
                path = os.path.normpath(path)
                if options.eject:
                    mounter.eject(path, force=True)
                if options.detach:
                    mounter.detach(path, force=True)
            else:
                device = mounter.unmount(os.path.normpath(path))
                if device:
                    unmounted.append(device)

    # automatically lock unused luks slaves of unmounted devices
    for device in unmounted:
        mounter.lock_slave(device)
    return 0




"""
Tests for the udiskie.common module.

Since the tested classes are mainly a very light wrapper around the
dbus-python API the tests are not very comprehensive. In the hope that
python-dbus is tested thoroughly, we need only check that the wiring works
at all.

"""
# test utility
import dbus
import dbusmock

# tested library:
from udiskie.common import DBusProxy


class TestDBusProxy(dbusmock.DBusTestCase):
    """
    Test that the udiskie.common.DBusProxy class is working correctly.

    More specifically the .method and the .property (class DBusProperties)
    members are tested to work as expected.

    """
    @classmethod
    def setUpClass(cls):
        cls.start_system_bus()
        cls.bus = cls.get_dbus(True)

    def setUp(self):
        self.srv = self.spawn_server('com.example.Simple',
                                     '/com/example/Simple',
                                     'com.example.Simple',
                                     system_bus=True)
        self.obj, = self.bus.get_object('com.example.Simple',
                                       '/com/example/Simple'),
        self.mock_prox = dbus.Interface(self.obj, dbusmock.MOCK_IFACE)
        self.mock_prox.AddMethod('', 'Square', 'i', 'i',
                                 'ret = args[0]*args[0]')
        self.mock_prox.AddProperty('', 'Foo', 'foo')
        self.mock_prox.AddProperty('', 'Bar', 'bar')
        self.prox = DBusProxy(self.obj, 'com.example.Simple')

    def tearDown(self):
        self.srv.terminate()
        self.srv.wait()

    def test_method_working(self):
        """Test that valid method call via the .method member will succeed."""
        self.assertEqual(self.prox.method.Square(3), 9)
        self.assertEqual(self.prox.method.Square(4), 16)

    def test_method_failing(self):
        """Test that invalid method call will result in correct exception."""
        try:
            self.prox.method.SurelyThisMethodDoesNotExistOnTheDBusObject()
        except self.prox.Exception:
            pass
        else:
            assert False

    def test_property_working(self):
        """Test that the attribute access works as expected."""
        self.assertEqual(str(self.prox.property.Foo), 'foo')
        self.assertEqual(str(self.prox.property.Bar), 'bar')

    def test_property_failing(self):
        """Test that invalid property access will result in exception."""
        try:
            self.prox.property.SurelyThisPropertyDoesNotExistOnTheDBusObject
        except self.prox.Exception:
            pass
        else:
            assert False




"""
Udiskie mount utilities.
"""
__all__ = ['Mounter']

import sys
import logging

class Mounter(object):
    """
    Mount utility.

    Remembers udisks, prompt and filter instances to use across multiple
    mount operations.

    """
    def __init__(self, filter=None, prompt=None, udisks=None):
        """
        Initialize mounter with the given defaults.

        The parameters are not guaranteed to keep their order and should
        always be passed as keyword arguments.

        If udisks is None only the xxx_device methods will work. The
        exception is the force=True branch of remove_device.

        If prompt is None device unlocking will not work.

        """
        self.filter = filter
        self.prompt = prompt
        self.udisks = udisks

    # mount/unmount
    def mount_device(self, device, filter=None):
        """
        Mount the device if not already mounted.

        Return value indicates whether an action was performed successfully.
        The special value `None` means unknown/unreliable.

        """
        log = logging.getLogger('udiskie.mount.mount_device')
        if not device.is_handleable or not device.is_filesystem:
            log.debug('skipping unhandled device %s' % (device,))
            return False
        if device.is_mounted:
            log.debug('skipping mounted device %s' % (device,))
            return False

        fstype = str(device.id_type)
        filter = filter or self.filter
        options = filter.get_mount_options(device) if filter else []

        log.info('attempting to mount device %s (%s:%s)' % (device, fstype, options))
        try:
            device.mount(fstype, options)
        except device.Exception:
            err = sys.exc_info()[1]
            log.error('failed to mount device %s: %s' % (device, err))
            return None

        log.info('mounted device %s' % (device,))
        return True

    def unmount_device(self, device):
        """
        Unmount a Device.

        Checks to make sure the device is unmountable and then unmounts.
        Return value indicates whether an action was performed successfully.
        The special value `None` means unknown/unreliable.

        """
        logger = logging.getLogger('udiskie.mount.unmount_device')
        if not device.is_handleable or not device.is_filesystem:
            logger.debug('skipping unhandled device %s' % (device,))
            return False
        if not device.is_mounted:
            logger.debug('skipping unmounted device %s' % (device,))
            return False
        try:
            device.unmount()
            logger.info('unmounted device %s' % (device,))
        except device.Exception:
            err = sys.exc_info()[1]
            logger.error('failed to unmount device %s: %s' % (device, err))
            return None
        return True

    # unlock/lock (LUKS)
    def unlock_device(self, device, prompt=None):
        """
        Unlock the device if not already unlocked.

        Return value indicates whether an action was performed successfully.
        The special value `None` means unknown/unreliable.

        """
        log = logging.getLogger('udiskie.mount.unlock_device')
        if not device.is_handleable or not device.is_crypto:
            log.debug('skipping unhandled device %s' % (device,))
            return False
        if device.is_unlocked:
            log.debug('skipping unlocked device %s' % (device,))
            return False

        # prompt user for password
        prompt = prompt or self.prompt
        password = prompt and prompt(
                'Enter password for %s:' % (device,),
                'Unlock encrypted device')
        if password is None:
            return False

        # unlock device
        log.info('attempting to unlock device %s' % (device,))
        try:
            device.unlock(password, [])
            holder_dev = device.create(
                device.bus,
                device.luks_cleartext_holder)
            holder_path = holder_dev.device_file
            log.info('unlocked device %s on %s' % (device, holder_path))
        except device.Exception:
            err = sys.exc_info()[1]
            log.error('failed to unlock device %s:\n%s' % (device, err))
            return None
        return True

    def lock_device(self, device):
        """
        Lock device.

        Checks to make sure the device is lockable, then locks.
        Return value indicates whether an action was performed successfully.
        The special value `None` means unknown/unreliable.

        """
        logger = logging.getLogger('udiskie.mount.lock_device')
        if not device.is_handleable or not device.is_crypto:
            logger.debug('skipping unhandled device %s' % (device,))
            return False
        if not device.is_unlocked:
            logger.debug('skipping locked device %s' % (device,))
            return False
        try:
            device.lock([])
            logger.info('locked device %s' % (device,))
        except device.Exception:
            err = sys.exc_info()[1]
            logger.error('failed to lock device %s: %s' % (device, err))
            return None
        return True


    # add/remove (unlock/lock or mount/unmount)
    def add_device(self, device, filter=None, prompt=None):
        """Mount or unlock the device depending on its type."""
        log = logging.getLogger('udiskie.mount.add_device')
        if not device.is_handleable:
            log.debug('skipping unhandled device %s' % (device,))
            return False
        if device.is_filesystem:
            return self.mount_device(device, filter)
        elif device.is_crypto:
            return self.unlock_device(device, prompt)

    def remove_device(self, device, force=False):
        """
        Unmount or lock the device depending on device type.

        If `force` is True recursively unmount/unlock all devices that are
        contained by this device.

        """
        logger = logging.getLogger('udiskie.mount.remove_device')
        if device.is_filesystem:
            return self.unmount_device(device)
        elif device.is_crypto:
            if not device.is_unlocked:
                return False
            if force:
                self.remove_device(
                    device.create(device.bus, device.luks_cleartext_holder),
                    force=True)
            return self.lock_device(device)
        elif force and device.is_partition_table:
            success = True
            for dev in self.udisks.get_all_handleable():
                if dev.is_partition and dev.partition_slave == device:
                    ret = self.remove_device(dev, force=True)
                    if ret is None:
                        success = None
                    else:
                        success = success and ret
            return success
        else:
            logger.debug('skipping unhandled device %s' % (device,))
            return False

    # eject/detach device
    def eject_device(self, device, force=False):
        """Eject a device after unmounting all its mounted filesystems."""
        logger = logging.getLogger('udiskie.mount.eject_device')
        drive = device.drive
        if drive.is_drive and drive.is_ejectable:
            try:
                if force:
                    drive.eject(['unmount'])
                else:
                    drive.eject([])
                logger.info('ejected device %s' % (device,))
                return True
            except drive.Exception:
                logger.warning('failed to eject device %s' % (device,))
                return False
        else:
            logger.debug('drive not ejectable: %s' % drive)
            return False

    def detach_device(self, device, force=False):
        """Detach a device after unmounting all its mounted filesystems."""
        logger = logging.getLogger('udiskie.mount.eject_device')
        drive = device.drive
        if drive.is_drive and drive.is_detachable:
            if force:
                self.remove_device(drive, force=True)
            try:
                drive.detach([])
                logger.info('detached device %s' % (device,))
                return True
            except drive.Exception:
                logger.warning('failed to detach device %s' % (device,))
                return False
        else:
            logger.debug('drive not detachable: %s' % drive)
            return False

    # mount_all/unmount_all
    def mount_all(self, filter=None, prompt=None):
        """Mount handleable devices that are already present."""
        for device in self.udisks.get_all_handleable():
            self.add_device(device, filter, prompt)

    def unmount_all(self):
        """Unmount all filesystems handleable by udiskie."""
        unmounted = []
        for device in self.udisks.get_all_handleable():
            if self.unmount_device(device):
                unmounted.append(device)
        return unmounted

    def eject_all(self):
        """Eject all ejectable devices."""
        ejected = []
        for device in self.udisks.get_all():
            if (device.is_drive and
                device.is_external and
                device.is_ejectable and
                eject_device(device, force=True)):
                ejected.append(device)
        return ejected

    def detach_all(self):
        """Detach all detachable devices."""
        detached = []
        for device in self.udisks.get_all():
            if (device.is_drive and
                device.is_external and
                device.is_detachable and
                self.detach_device(device, force=True)):
                detached.append(device)
        return detached

    # mount a holder/lock a slave
    def mount_holder(self, device, filter=None, prompt=None):
        """
        Mount or unlock the holder device of this unlocked LUKS device.

        Will not mount the holder if the device is not unlocked.
        Return value indicates success

        """
        logger = logging.getLogger('udiskie.mount.lock_slave')
        if not device.is_unlocked:
            logger.debug('skipping locked or non-luks device %s' % (device,))
            return False
        holder_path = device.luks_cleartext_holder
        holder = self.udisks.create_device(holder_path)
        return self.add_device(holder, filter=filter, prompt=prompt)

    def lock_slave(self, device):
        """
        Lock the luks slave of this device.

        Will not lock the slave if it is still used by any mounted file system.
        Return value indicates success.

        """
        logger = logging.getLogger('udiskie.mount.lock_slave')
        if not device.is_luks_cleartext:
            logger.debug('skipping non-luks-cleartext device %s' % (device,))
            return False
        slave_path = device.luks_cleartext_slave
        slave = self.udisks.create_device(slave_path)
        if slave.is_luks_cleartext_slave:
            return False
        return self.lock_device(slave)


    # mount/unmount by path
    def mount(self, path, filter=None, prompt=None):
        """
        Mount or unlock a device.

        The device must match the criteria for a filesystem mountable or
        unlockable by udiskie.

        """
        logger = logging.getLogger('udiskie.mount.unmount')
        device = self.udisks.get_device(path)
        if device:
            logger.debug('found device owning "%s": "%s"' % (path, device))
            if self.add_device(device, filter=filter, prompt=prompt):
                return device
        return None

    def unmount(self, path, force=False):
        """
        Unmount or lock a filesystem

        The filesystem must match the criteria for a filesystem mountable by
        udiskie.  path is either the physical device node (e.g. /dev/sdb1) or the
        mount point (e.g. /media/Foo).

        """
        logger = logging.getLogger('udiskie.mount.unmount')
        device = self.udisks.get_device(path)
        if device:
            logger.debug('found device owning "%s": "%s"' % (path, device))
            if self.remove_device(device, force=force):
                return device
        return None

    # eject media/detach drive
    def eject(self, path, force=False):
        """
        Eject media from the device.

        If the device has any mounted filesystems, these will be unmounted
        before ejection.

        """
        logger = logging.getLogger('udiskie.mount.eject')
        device = self.udisks.get_device(path)
        if device:
            logger.debug('found device owning "%s": "%s"' % (path, device))
            if self.eject_device(device, force=force):
                return device
        else:
            logger.warning('no found device owning "%s"' % (path))
            return None

    def detach(self, path, force=False):
        """
        Eject media from the device.

        If the device has any mounted filesystems, these will be unmounted
        before ejection.

        """
        logger = logging.getLogger('udiskie.mount.detach')
        device = self.udisks.get_device(path)
        if device:
            logger.debug('found device owning "%s": "%s"' % (path, device))
            if self.detach_device(device, force=force):
                return device
        else:
            logger.warning('no found device owning "%s"' % (path))
            return None



# encoding: utf-8
from setuptools import setup

# check availability of runtime dependencies
def check_any(*packages):
    """Issue a warning if none of the packages is available."""
    errors = []
    for package in packages:
        try:
            __import__(package)
            return True
        except ImportError:
            import sys
            errors.append(sys.exc_info()[1])
    if len(errors) == 1:
        print("Missing runtime dependency: %s" % errors[0])
    else:
        print("Missing runtime dependencies:")
        for err in errors:
            print("\t%s" % err)
    return False

check_any('dbus')
check_any('gobject')
check_any('pynotify', 'notify2')
check_any('gtk')

# read long_description from README.rst
try:
    f = open('README.rst')
    long_description = f.read()
    f.close()
except IOError:
    long_description = None

setup(
    name='udiskie',
    version='0.5.2',
    description='Removable disk automounter for udisks',
    long_description=long_description,
    author='Byron Clark',
    author_email='byron@theclarkfamily.name',
    maintainer='Thomas Gläßle',
    maintainer_email='t_glaessle@gmx.de',
    url='https://github.com/coldfix/udiskie',
    license='MIT',
    packages=[
        'udiskie',
    ],
    namespace_packages=[
        'udiskie',
    ],
    entry_points={
        'console_scripts': [
            'udiskie = udiskie.cli:daemon',
            'udiskie-mount = udiskie.cli:mount',
            'udiskie-umount = udiskie.cli:umount',
        ],
    },
    extras_require={
        'notifications': ['notify2']
    },
    install_requires=[
        # Currently not building out of the box:
        # 'PyGObject',
        # 'dbus-python',
        # 'pygtk>=2.10',
    ],
    tests_require=[
        'python-dbusmock>=0.7.2'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: X11 Applications',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Topic :: Desktop Environment',
        'Topic :: Software Development',
        'Topic :: System :: Filesystems',
        'Topic :: System :: Hardware',
        'Topic :: Utilities',
    ],
)



"""
Udisks event daemon module.

Provides the class `Daemon` which listens to udisks events. When a change
occurs this class detects what has changed and triggers an appropriate
event.

"""
__all__ = ['Daemon']

import logging
import sys


class DeviceState(object):
    """
    State information struct for devices.
    """
    __slots__ = ['mounted', 'has_media', 'unlocked']

    def __init__(self, mounted, has_media, unlocked):
        self.mounted = mounted
        self.has_media = has_media
        self.unlocked = unlocked

class Daemon(object):
    """
    Udisks listener daemon.

    Listens to udisks events. When a change occurs this class detects what
    has changed and triggers an appropriate event. Valid events are:

        - device_added    / device_removed
        - device_unlocked / device_locked
        - device_mounted  / device_unmounted
        - media_added     / media_removed
        - device_changed

    A very primitive mechanism that gets along without external
    dependencies is used for event dispatching. The methods `connect` and
    `disconnect` can be used to add or remove event handlers.

    """
    def __init__(self, udisks):
        """
        Initialize object and start listening to udisks events.
        """
        self.log = logging.getLogger('udiskie.daemon.Daemon')
        self.state = {}
        self.udisks = udisks

        self.event_handlers = {
            'device_added': [],
            'device_removed': [],
            'device_mounted': [],
            'device_unmounted': [],
            'media_added': [],
            'media_removed': [],
            'device_unlocked': [],
            'device_locked': [],
            'device_changed': [self.on_device_changed]
        }

        for device in self.udisks.get_all_handleable():
            self._store_device_state(device)

        udisks.bus.add_signal_receiver(
                self._device_added,
                signal_name='DeviceAdded',
                bus_name='org.freedesktop.UDisks')
        udisks.bus.add_signal_receiver(
                self._device_removed,
                signal_name='DeviceRemoved',
                bus_name='org.freedesktop.UDisks')
        udisks.bus.add_signal_receiver(
                self._device_changed,
                signal_name='DeviceChanged',
                bus_name='org.freedesktop.UDisks')

    # events
    def on_device_changed(self, udevice, old_state, new_state):
        """Detect type of event and trigger appropriate event handlers."""
        if old_state is None:
            self.trigger('device_added', udevice)
            return
        d = {}
        d['device_mounted'] = new_state.mounted and not old_state.mounted
        d['device_unmounted'] = old_state.mounted and not new_state.mounted
        d['media_added'] = new_state.has_media and not old_state.has_media
        d['media_removed'] = old_state.has_media and not new_state.has_media
        d['device_unlocked'] = new_state.unlocked and not old_state.unlocked
        d['device_locked'] = old_state.unlocked and not new_state.unlocked
        for event in d:
            if d[event]:
                self.trigger(event, udevice)

    # event machinery
    def trigger(self, event, device, *args):
        """Trigger event handlers."""
        self.log.debug('%s: %s' % (event, device))
        for handler in self.event_handlers[event]:
            handler(device, *args)

    def connect(self, handler, event=None):
        """Connect an event handler."""
        if event:
            self.event_handlers[event].append(handler)
        else:
            for event in self.event_handlers:
                if hasattr(handler, event):
                    self.connect(getattr(handler, event), event)

    def disconnect(self, handler, event=None):
        """Disconnect an event handler."""
        if event:
            self.event_handlers.remove(handler)
        else:
            for event in self.event_handlers:
                if hasattr(handler, event):
                    self.disconnect(getattr(handler, event), event)

    # udisks event listeners
    def _device_added(self, object_path):
        try:
            udevice = self.udisks.create_device(object_path)
            if not udevice.is_handleable:
                return
            self._store_device_state(udevice)
            self.trigger('device_added', udevice)
        except self.udisks.Exception:
            err = sys.exc_info()[1]
            self.log.error('%s(%s): %s' % ('device_added', object_path, err))

    def _device_removed(self, object_path):
        try:
            self.trigger('device_removed', object_path)
            self._remove_device_state(object_path)
        except self.udisks.Exception:
            err = sys.exc_info()[1]
            self.log.error('%s(%s): %s' % ('device_removed', object_path, err))

    def _device_changed(self, object_path):
        try:
            udevice = self.udisks.create_device(object_path)
            if not udevice.is_handleable:
                return
            old_state = self._get_device_state(object_path)
            new_state = self._store_device_state(udevice)
            self.trigger('device_changed', udevice, old_state, new_state)
        except self.udisks.Exception:
            err = sys.exc_info()[1]
            self.log.error('%s(%s): %s' % ('device_changed', object_path, err))

    # internal state keeping
    def _store_device_state(self, device):
        self.state[device.object_path] = DeviceState(
            device.is_mounted,
            device.has_media,
            device.is_unlocked)
        return self.state[device.object_path]

    def _remove_device_state(self, object_path):
        if object_path in self.state:
            del self.state[object_path]

    def _get_device_state(self, object_path):
        return self.state.get(object_path)




"""
Filters for udiskie mount tools.
"""
__all__ = ['InvalidFilter', 'OptionFilter', 'Filters', 'FilterMatcher']

try:
    from ConfigParser import SafeConfigParser
except ImportError:
    from configparser import SafeConfigParser

import logging
import re

class InvalidFilter(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return 'Invalid Filter: %s' % (self.value,)


class OptionFilter(object):
    # This list also defines the order in which the filters are
    # processed. Should go from least specific to most specific.
    VALID_PARAMETERS = (
        'fstype',
        'uuid',
    )

    MATCH_PATTERN = re.compile(r'(\w+)\.(\S+)')

    def __init__(self, parameter, value, options):
        self.parameter = parameter
        self.value = value
        self.options = options

    def __str__(self):
        return '<OptionFilter: %s=%s, options=%s>' % (self.parameter,
                                                      self.value,
                                                      self.options)

    def __repr__(self):
        return str(self)


class Filters(object):
    def __init__(self):
        self.option_filters = []
        self.log = logging.getLogger('udiskie.match.Filters')

    def _parse_option_match(self, match_expression):
        match = OptionFilter.MATCH_PATTERN.match(match_expression)
        if not match:
            raise InvalidFilter('format is parameter=value')
        parameter, value = match.groups()
        if parameter not in OptionFilter.VALID_PARAMETERS:
            raise InvalidFilter('parameter "%s" is not allowed' % (parameter,))
        return parameter, value

    def add_option_filter(self, match_expression, options):
        parameter, value = self._parse_option_match(match_expression)
        options = [S.strip() for S in options.split(',')]
        filt = OptionFilter(parameter, value, options)
        self.log.debug('loaded filter: %s' % (filt,))
        self.option_filters.append(filt)

    def get_option_filters(self, parameter):
        return [F for F in self.option_filters if F.parameter == parameter]


class FilterMatcher(object):
    MOUNT_OPTIONS_SECTION = 'mount_options'

    def __init__(self, config_files):
        self.log = logging.getLogger('udiskie.match.FilterMatcher')
        self.filters = self._load_filters_from_config_files(config_files)

    def _load_filters_from_config_files(self, config_files):
        filters = Filters()

        self.log.debug('loading filters from %s' % (config_files,))
        parser = SafeConfigParser()
        parser.read(config_files)

        if parser.has_section(self.MOUNT_OPTIONS_SECTION):
            for (match, options) in parser.items(self.MOUNT_OPTIONS_SECTION):
                filters.add_option_filter(match, options)

        return filters

    def get_mount_options(self, device):
        device_info = {
            'fstype' : device.id_type,
            'uuid' : device.id_uuid.lower()
        }

        mount_options = set()
        for match_type in OptionFilter.VALID_PARAMETERS:
            device_value = device_info.get(match_type)
            for filt in self.filters.get_option_filters(match_type):
                if device_value == filt.value:
                    self.log.info('filter matched: %s' % (filt,))
                    for option in filt.options:
                        mount_options.add(option)

        return list(mount_options)
